from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import time

from .exceptions import MissingStepError
from .utils import to_docstring


class IVisitor(object):

    __metaclass__ = ABCMeta

    def before_feature(self, node):
        pass

    def after_feature(self, node):
        pass

    def before_scenario(self, node):
        pass

    def after_scenario(self, node):
        pass

    def permute_schedule(self, node):
        return [[0]]

    def step_schedule(self, node):
        return [list(range(len(node.steps)))]

    @abstractmethod
    def visit(self):
        pass  # pragma: nocover


class TestVisitor(IVisitor):
    """ Visits all steps and run step methods. """

    def __init__(self, suite, matcher, formatter):
        self._suite = suite
        self._matcher = matcher
        self._formatter = formatter
        self._scenarios_num = 0
        self._steps_num = 0

    def visit(self, node):
        self._suite.step = node
        self._steps_num += 1
        line = node.get_real_reconstruction(self._suite, self._matcher)
        start_time = time.time()
        try:
            try:
                node.test_step(self._suite, self._matcher)
            except (MissingStepError, AssertionError):
                status = 'fail'
                raise
            except (SystemExit, Exception):
                status = 'error'
                raise
            else:
                status = 'pass'
        finally:
            end_time = time.time()
            duration = end_time - start_time
            self._formatter.output(node, line, status, duration)

    def before_scenario(self, node):
        self._suite.setUp()
        self._scenarios_num += 1

    def after_scenario(self, node):
        self._suite.tearDown()

    def permute_schedule(self, node):
        return node.permute_schedule()

    def step_schedule(self, node):
        return node.step_schedule()


class ReportVisitor(IVisitor):
    def __init__(self, suite):
        self._suite = suite
        self._result = ''

    def visit(self, node):
        recon, result = node.to_html()
        if recon[-1] != '\n':
            recon += '\n'  # TODO  clean this outa def reconstruction(s)!
        self._result += recon
        return result

    def after_scenario(self, node):
        if node.result:
            self._result += node.result

    def __str__(self):
        return self._result

    def __unicode__(self):
        return unicode(self._result)


class StepMatcherVisitor(IVisitor):
    """ Visits all steps in order to find missing step methods. """

    def __init__(self, suite, matcher):
        self._suite = suite
        self._not_matched = OrderedDict()
        self._matcher = matcher

    def visit(self, node):
        try:
            node.find_step(self._suite, self._matcher)
        except MissingStepError as e:
            if e.docstring:
                self._not_matched[e.docstring] = e.suggest
            else:
                self._not_matched[e.method_name] = e.suggest

    def after_feature(self, node):
        suggest = u''.join(self._not_matched.values())
        if suggest:
            diagnostic = u'Cannot match steps:\n\n%s' % suggest
            self._suite.fail(to_docstring(diagnostic))



from abc import ABCMeta
import copy
import itertools
import re

from six import moves

from .exceptions import MissingStepError
from .formatters import NullFormatter
from .i18n import TRANSLATIONS
from .matchers import RegexpStepMatcher, ParseStepMatcher, MethodNameStepMatcher
from .utils import fix_exception_encoding, to_unicode
from .visitors import StepMatcherVisitor, TestVisitor, ReportVisitor


class AST(object):

    def __init__(self, steps, test_visitor_class=TestVisitor,
                 matcher_visitor_class=StepMatcherVisitor):
        self.steps = steps
        self._test_visitor_class = test_visitor_class
        self._matcher_visitor_class = matcher_visitor_class

    def evaluate(self, suite, formatter=None, matchers=None, show_all_missing=False):
        if matchers is None:
            matchers = [RegexpStepMatcher, ParseStepMatcher, MethodNameStepMatcher]
        matcher = self._create_matchers_chain(suite, matchers)
        feature = self.steps[0]
        if show_all_missing:
            matcher_visitor = self._matcher_visitor_class(suite, matcher)
            feature.evaluate_steps(matcher_visitor)
        if formatter is None:
            formatter = NullFormatter()
        test_visitor = self._test_visitor_class(suite, matcher, formatter)
        try:
            feature.evaluate_steps(test_visitor)
        except SyntaxError as exc:
            raise
        except Exception as exc:
            fix_exception_encoding(exc)
            raise

    def _create_matchers_chain(self, suite, matcher_classes):
        root_matcher = None
        for matcher_class in matcher_classes:
            matcher = matcher_class(suite)
            try:
                root_matcher.add_matcher(matcher)
            except AttributeError:
                root_matcher = matcher
        return root_matcher

    def report(self, suite, visitor_class=ReportVisitor):
        rv = visitor_class(suite)
        self.steps[0].evaluate_steps(rv)
        return rv


class INode(object):

    __metaclass__ = ABCMeta

    def test_step(self, suite, matcher):
        pass

    def find_step(self, suite, matcher):
        pass

    def evaluate_steps(self, visitor):
        class_name = self.__class__.__name__.lower()
        self._method_hook(visitor, class_name, 'before_')
        try:
            visitor.visit(self)
            self._evaluate_child_steps(visitor)
        finally:
            self._method_hook(visitor, class_name, 'after_')

    def _evaluate_child_steps(self, visitor):
        for step in self.steps:
            step.evaluate_steps(visitor)

    def _method_hook(self, visitor, class_name, prefix):
        method = getattr(visitor, 'after_%s' % class_name, None)
        if method:
            method(self)


class Morelia(INode):

    def __init__(self, keyword, predicate):
        self.parent = None
        self.additional_data = {}
        self.keyword = keyword
        self.predicate = predicate if predicate is not None else ''

    def is_executable(self):
        return False

    def connect_to_parent(self, steps=[], line_number=0):
        self.steps = []
        self.line_number = line_number

#  TODO  escape the sample regices already!
#  and the default code should be 'print <arg_names, ... >'

        mpt = self.my_parent_type()
        try:
            for step in steps[::-1]:
                if isinstance(step, mpt):
                    step.steps.append(self)  # TODO  squeek if can't find parent
                    self.parent = step
                    break
        except TypeError:
            self.enforce(False, 'Only one Feature per file')  # CONSIDER  prevent it don't trap it!!!

        steps.append(self)
        return self

    def prefix(self):
        return ''

    def my_parent_type(self):
        return None

    @classmethod
    def get_pattern(cls, language):
        class_name = cls.__name__
        name = class_name.lower()
        name = TRANSLATIONS[language].get(name, class_name)
        return r'\s*(?P<keyword>' + name + '):?(?:\s+(?P<predicate>.*))?$'

    def count_dimensions(self):
        ''' Get number of rows. '''
        return sum([step.count_dimension() for step in self.steps])

    def count_dimension(self):  # CONSIDER  beautify this crud!
        return 0

    def validate_predicate(self):
        return  # looks good! (-:

    def enforce(self, condition, diagnostic):
        if not condition:
            text = ''
            offset = 1
            if self.parent:
                text = self.parent.reconstruction()
                offset = 5
            text += self.reconstruction()
            text = text.replace('\n\n', '\n').replace('\n', '\n\t')
            raise SyntaxError(diagnostic, (self.get_filename(), self.line_number, offset, text))

    def format_fault(self, diagnostic):
        parent_reconstruction = ''
        if self.parent:
            parent_reconstruction = self.parent.reconstruction().strip('\n')
        reconstruction = self.reconstruction()
        args = (self.get_filename(), self.line_number, parent_reconstruction, reconstruction, diagnostic)
        args = tuple([to_unicode(i) for i in args])
        return u'\n  File "%s", line %s, in %s\n %s\n%s' % args

    def reconstruction(self):
        predicate = to_unicode(self.predicate)
        recon = u'%s%s: %s' % (self.prefix(), self.keyword, predicate)
        if recon[-1] != u'\n':
            recon += u'\n'
        return recon

    def get_real_reconstruction(self, suite, macher):
        return self.reconstruction() + '\n'

    def get_filename(self):
        node = self

        while node:
            if not node.parent and hasattr(node, 'filename'):
                return node.filename
            node = node.parent

        return None


class Feature(Morelia):

    def test_step(self, suite, matcher):
        self.enforce(0 < len(self.steps), 'Feature without Scenario(s)')

    def to_html(self):
        return ['''\n<div><table>
                <tr style="background-color: #aaffbb;" width="100%%">
                <td align="right" valign="top" width="100"><em>%s</em>:</td>
                <td colspan="101">%s</td>
                </tr></table></div>''' % (self.keyword, _clean_html(self.predicate)), '']

    def prepend_steps(self, scenario):
        background = scenario.parent.steps[0]
        try:
            background.prepend_steps(scenario)
        except AttributeError:
            pass

    def reconstruction(self):
        predicate = to_unicode(self.predicate)
        predicate = predicate.replace('\n', '\n    ')
        recon = u'%s%s: %s' % (self.prefix(), self.keyword, predicate)
        if recon[-1] != u'\n':
            recon += u'\n'
        return recon


class Scenario(Morelia):

    def my_parent_type(self):
        return Feature

    def evaluate_steps(self, visitor):
        if self.parent:
            self.parent.prepend_steps(self)
        step_schedule = visitor.step_schedule(self)  # TODO  test this permuter directly (and rename it already)

        for step_indices in step_schedule:   # TODO  think of a way to TDD this C-:
            schedule = visitor.permute_schedule(self)

            for indices in schedule:
                self.row_indices = indices
                self.evaluate_test_case(visitor, step_indices)  # note this works on reports too!

    def evaluate_test_case(self, visitor, step_indices):  # note this permutes reports too!
        self.enforce(0 < len(self.steps), 'Scenario without step(s) - Step, Given, When, Then, And, or #')

        visitor.before_scenario(self)
        try:
            self.result = visitor.visit(self)

            for idx, step in enumerate(self.steps):
                if idx in step_indices:
                    step.evaluate_steps(visitor)
        finally:
            visitor.after_scenario(self)

    def permute_schedule(self):  # TODO  rename to permute_row_schedule
        dims = self.count_Row_dimensions()
        return _permute_indices(dims)

    def step_schedule(self):  # TODO  rename to permute_step_schedule !
        sched = []
        pre_slug = []

        #  TODO   deal with steps w/o whens

        for idx, s in enumerate(self.steps):
            if s.__class__ == When:
                break
            else:
                pre_slug.append(idx)

        for idx, s in enumerate(self.steps):
            if s.__class__ == When:
                slug = pre_slug[:]
                slug.append(idx)

                for idx in range(idx + 1, len(self.steps)):
                    s = self.steps[idx]
                    if s.__class__ == When:
                        break
                    slug.append(idx)

                sched.append(slug)

        if sched == []:
            return [pre_slug]
        return sched

    def count_Row_dimensions(self):
        return [step.count_dimensions() for step in self.steps]

    def reconstruction(self):
        return '\n' + self.keyword + ': ' + self.predicate

    def to_html(self):
        return ['''\n<div><table width="100%%">
                <tr style="background-color: #cdffb8;">
                <td align="right" valign="top" width="100"><em>%s</em>:</td>
                <td colspan="101">%s</td></tr>''' % (self.keyword, _clean_html(self.predicate)),
                '</table></div>']


class Background(Morelia):

    def my_parent_type(self):
        return Feature

    def _evaluate_child_steps(self, visitor):
        pass

    def prepend_steps(self, scenario):
        new_steps = []
        for step in self.steps:
            new_step = copy.copy(step)
            new_step.parent = scenario
            new_steps.append(new_step)
        scenario.steps = new_steps + scenario.steps


class Step(Morelia):

    def prefix(self):
        return '  '

    def is_executable(self):
        return True

    def my_parent_type(self):
        return (Scenario, Background)

    def find_step(self, suite, matcher):
        predicate = self.predicate
        augmented_predicate = self._augment_predicate()
        method, args, kwargs = matcher.find(predicate, augmented_predicate)
        if method:
            return method, args, kwargs

        suggest, method_name, docstring = matcher.suggest(predicate)
        raise MissingStepError(predicate, suggest, method_name, docstring)

    def get_real_reconstruction(self, suite, matcher):
        predicate = to_unicode(self._augment_predicate())
        recon = u'    %s %s' % (self.keyword, predicate)
        if recon[-1] != u'\n':
            recon += u'\n'
        return recon

    def _augment_predicate(self):  # CONSIDER  unsucktacularize me pleeeeeeze
        if self.parent is None:
            return self.predicate
        dims = self.parent.count_Row_dimensions()
        if set(dims) == set([0]):
            return self.predicate
        rep = re.compile(r'\<(\w+)\>')
        replitrons = rep.findall(self.predicate)
        if replitrons == []:
            return self.predicate
        self.copy = self.predicate[:]
        row_indices = self.parent.row_indices

        for replitron in replitrons:
            for x in range(0, len(row_indices)):
                table = self.parent.steps[x].steps

                try:
                    row = next(moves.filter(lambda step: isinstance(step, Row), table))
                except StopIteration:
                    pass
                else:
                    for q, title in enumerate(row.harvest()):
                        self.replace_replitron(x, q, row_indices, table, title, replitron)

        return self.copy

    def evaluate(self, suite, matcher):
        method, args, kwargs = self.find_step(suite, matcher)
        method(*args, **kwargs)

    def test_step(self, suite, matcher):
        try:
            self.evaluate(suite, matcher)
        except MissingStepError as exc:
            message = self.format_fault(exc.suggest)
            exc.args = (message,)
            raise
        except Exception as exc:
            if len(exc.args):
                message = exc.args[0]
                message = self.format_fault(message)
                exc.args = (message,) + exc.args[1:]
            raise

    def replace_replitron(self, x, q, row_indices, table, title, replitron):
        if title != replitron:
            return
        at = row_indices[x] + 1

        assert at < len(table), 'CONSIDER this should never happen'

        #  CONSIDER  we hit this too many times - hit once and stash the result
        #  CONSIDER  better diagnostics when we miss these

        stick = table[at].harvest()
        found = stick[q]  # CONSIDER  this array overrun is what you get when your table is ragged
        # CONSIDER  only if it's not nothing?
        found = found.replace('\n', '\\n')  # CONSIDER  crack the multi-line argument bug, and take this hack out!
        self.copy = self.copy.replace('<%s>' % replitron, found)

        # CONSIDER  mix replitrons and matchers!

    def to_html(self):
        return '\n<tr><td align="right" valign="top"><em>' + self.keyword + '</em></td><td colspan="101">' + _clean_html(self.predicate) + '</td></tr>', ''


class Given(Step):  # CONSIDER  distinguish these by fault signatures!

    pass


class When(Step):  # TODO  cycle these against the Scenario

    def to_html(self):
        return '\n<tr style="background-color: #cdffb8; background: url(http://www.zeroplayer.com/images/stuff/aqua_gradient.png) no-repeat; background-size: 100%;"><td align="right" valign="top"><em>' + self.keyword + '</em></td><td colspan="101">' + _clean_html(self.predicate) + '</td></tr>', ''


class Then(Step):

    pass


class And(Step):

    pass


class But(And):

    pass

#  CONSIDER  how to validate that every row you think you wrote actually ran?


class Row(Morelia):

    @classmethod
    def get_pattern(cls, language):
        return r'\s*(?P<keyword>\|):?\s+(?P<predicate>.*)'

    def my_parent_type(self):
        return (Step, Examples)

    def prefix(self):
        return ' ' * 8

    def reconstruction(self):  # TODO  strip the reconstruction at error time
        recon = self.prefix() + '| ' + self.predicate
        if recon[-1] != '\n':
            recon += '\n'
        return recon

    def to_html(self):
        html = '\n<tr><td></td>'
        idx = self.parent.steps.index(self)
        em = 'span'
        if idx == 0:
            color = 'silver'
            em = 'em'
        elif ((2 + idx) / 3) % 2 == 0:
            color = '#eeffff'
        else:
            color = '#ffffee'

        for col in self.harvest():
            html += '<td style="background-color: %s;"><%s>' % (color, em) + _clean_html(col) + '</%s></td>' % em

        html += '<td>&#160;</td></tr>'  # CONSIDER  the table needn't stretch out so!
        return html, ''

    def count_dimension(self):
        if self is self.parent.steps[0]:
            # header row
            return 0
        return 1  # TODO  raise an error (if the table has one row!)

    def harvest(self):
        row = re.split(r' \|', re.sub(r'\|$', '', self.predicate))
        row = [s.strip() for s in row]
        return row


class Examples(Morelia):

    def prefix(self):
        return ' ' * 4

    def my_parent_type(self):
        return Scenario

#  TODO  sample data with "post-it haiku"
#  CONSIDER  trailing comments


class Comment(Morelia):

    def my_parent_type(self):
        return Morelia  # aka "any"

    @classmethod
    def get_pattern(cls, language):
        return r'\s*(?P<keyword>\#)(?P<predicate>.*)'

    def validate_predicate(self):
        self.enforce(self.predicate.count('\n') == 0, 'linefeed in comment')

    def reconstruction(self):
        recon = '    # ' + self.predicate
        if recon[-1] != '\n':
            recon += '\n'
        return recon

    def to_html(self):
        return '\n# <em>' + _clean_html(self.predicate) + '</em><br/>', ''


def _special_range(n):  # CONSIDER  better name
    return moves.range(n) if n else [0]


def _permute_indices(arr):
    product_args = list(_imap(arr))
    result = list(itertools.product(*product_args))
    return result
    #  tx to Chris Rebert, et al, on the Python newsgroup for curing my brainlock here!!


def _imap(*iterables):
    iterables = [iter(i) for i in iterables]
    while True:
        args = [next(i) for i in iterables]
        yield _special_range(*args)


def _clean_html(string):
    return string.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;').replace("'", '&#39;')



from abc import ABCMeta, abstractmethod
import sys

colors = {
    'normal': u'\x1b[30m',
    'fail': u'\x1b[31m',
    'error': u'\x1b[31m',
    'pass': u'\x1b[32m',
    'reset': u'\x1b[0m',
}


class IFormatter(object):

    ''' Abstract Base Class for all formatters. '''

    __metaclass__ = ABCMeta

    @abstractmethod
    def output(self, node, line, status, duration):
        ''' Method called after execution each step.

        :param INode node: node representing step
        :param str line: text of executed step
        :param str status: execution status
        :param float duration: step execution duration
        '''
        pass  # pragma: nocover


class NullFormatter(IFormatter):

    ''' Formatter that... do nothing. '''

    def output(self, node, line, status, duration):
        ''' See :py:meth:`IFormatter.output` '''

        pass


class PlainTextFormatter(IFormatter):

    ''' Formatter that prints all executed steps in plain text to a given stream. '''

    def __init__(self, stream=None):
        ''' Initialize formatter.

        :param file stream: file-like stream to output executed steps
        '''
        self._stream = stream if stream is not None else sys.stderr

    def output(self, node, line, status, duration):
        ''' See :py:meth:`IFormatter.output` '''

        if node.is_executable():
            status = status.lower()
            text = '%-60s # %-5s %.3fs\n' % (
                line.strip('\n'),
                status,
                duration,
            )
        else:
            text = '%s\n' % line.strip('\n')
        self._stream.write(text)
        self._stream.flush()


class ColorTextFormatter(PlainTextFormatter):

    ''' Formatter that prints all executed steps in color to a given stream. '''

    def output(self, node, line, status, duration):
        ''' See :py:meth:`IFormatter.output` '''

        if node.is_executable():
            status = status.lower()
            text = '%s%-60s # %.3fs%s\n' % (
                colors[status],
                line.strip('\n'),
                duration,
                colors['reset']
            )
        else:
            text = '%s\n' % line.strip('\n')
        self._stream.write(text)
        self._stream.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase

from mock import sentinel, Mock, patch

from morelia import run
from morelia.parser import Parser, DEFAULT_LANGUAGE
from morelia.grammar import Morelia, Step
from morelia.exceptions import MissingStepError


class ParserParseLanguageDirectiveTestCase(TestCase):
    """ Test :py:meth:`Parser.parse_language_directive`. """

    def test_should_return_true_on_language_directive_line(self):
        """ Scenario: line with language directive """
        # Arrange
        obj = Parser()
        for lang in ['en', 'pl', 'ja']:
            line = '# language: %s' % lang
            # Act
            result = obj._parse_language_directive(line)
            # Assert
            self.assertTrue(result)
            self.assertEqual(obj.language, lang)

    def test_should_return_false_if_no_language_directive(self):
        """ Scenario: line without language directive """
        # Arrange
        obj = Parser()
        lines = [
            '# language: ',  # missing language
            'Feature:',  # not a language directive
            'language: pl',  # missing comment
            '# comment',  # comment
        ]
        for line in lines:
            # Act
            result = obj._parse_language_directive(line)
            # Assert
            self.assertFalse(result)
            self.assertEqual(obj.language, DEFAULT_LANGUAGE)


class MoreliaReconstructionTestCase(TestCase):
    """ Test :py:meth:`Morelia.reconstruction`. """

    def test_should_reconstruct_unicode(self):
        """ Scenario: unicode input """
        # Arrange
        obj = Morelia('???', u'zażółć gęślą jaźń')
        # Act
        result = obj.reconstruction()
        # Assert
        self.assertEqual(result, u'???: zażółć gęślą jaźń\n')

    def test_should_reconstruct_utf8(self):
        """ Scenario: utf8 input """
        # Arrange
        obj = Morelia('???', 'zażółć gęślą jaźń')
        # Act
        result = obj.reconstruction()
        # Assert
        self.assertEqual(result, u'???: zażółć gęślą jaźń\n')


class MissingStepErrorTestCase(TestCase):
    """ Test :py:meth:`MissingStepError`. """

    def test_should_create_exception(self):
        """ Scenario: exception with step and suggest """
        # Arrange
        # Act
        result = MissingStepError('predicate line', 'def step_predicate_line(self):\n    pass', 'step_predicate_line', '')
        # Assert
        self.assertTrue(result is not None)
        self.assertEqual(result.predicate, 'predicate line')


class StepFindStepTestCase(TestCase):
    """ Test :py:meth:`Step.find_step`. """

    def test_should_find_method(self):
        """ Scenario: method found """
        # Arrange
        obj = Step('???', 'method')
        # Act
        matcher = Mock()
        matcher.find.return_value = (sentinel.method, [], {})
        result, args, kwargs = obj.find_step(sentinel.suite, matcher)
        # Assert
        self.assertEqual(result, sentinel.method)

    def test_should_not_find_method(self):
        """ Scenario: not found """
        # Arrange
        obj = Step('???', u'some_method')
        matcher = Mock()
        matcher.find.return_value = (None, [], {})
        matcher.suggest.return_value = ('suggest', 'method_name', 'docstring')
        self.assertRaises(MissingStepError, obj.find_step, sentinel.suite, matcher)


class RunTestCase(TestCase):
    """ Test :py:func:`run`. """

    @patch.object(Parser, 'parse_file')
    def test_should_parse_file_and_evaluate_tests(self, parse_file):
        """ Scenario: run and evaluate """
        run(sentinel.filename, sentinel.suite)
        parse_file.assert_called_once_with(sentinel.filename)
        parse_file.return_value.evaluate.assert_called_once_with(sentinel.suite)

    @patch('morelia.has_color_support')
    @patch('morelia.PlainTextFormatter')
    @patch.object(Parser, 'parse_file')
    def test_should_run_verbose_with_plain_text_formatter(self, parse_file, PlainTextFormatter, has_color_support):
        """ Scenario: run verbose on windows """
        has_color_support.return_value = False
        run(sentinel.filename, sentinel.suite, verbose=True)
        parse_file.assert_called_once_with(sentinel.filename)
        parse_file.return_value.evaluate.assert_called_once_with(sentinel.suite, formatter=PlainTextFormatter.return_value)

    @patch('morelia.has_color_support')
    @patch('morelia.ColorTextFormatter')
    @patch.object(Parser, 'parse_file')
    def test_should_run_verbose_with_color_text_formatter(self, parse_file, ColorTextFormatter, has_color_support):
        """ Scenario: run verbose on systems with color support """
        has_color_support.return_value = True
        run(sentinel.filename, sentinel.suite, verbose=True)
        parse_file.assert_called_once_with(sentinel.filename)
        parse_file.return_value.evaluate.assert_called_once_with(sentinel.suite, formatter=ColorTextFormatter.return_value)

    @patch.object(Parser, 'parse_file')
    def test_should_use_provided_formatter(self, parse_file):
        """ Scenario: run verbose on systems with color support """
        run(sentinel.filename, sentinel.suite, verbose=True, formatter=sentinel.formatter)
        parse_file.assert_called_once_with(sentinel.filename)
        parse_file.return_value.evaluate.assert_called_once_with(sentinel.suite, formatter=sentinel.formatter)



# -*- coding: utf-8 -*-
#            __  __                _ _
#           |  \/  | ___  _ __ ___| (_) __ _
#           | |\/| |/ _ \| '__/ _ \ | |/ _` |
#           | |  | | (_) | | |  __/ | | (_| |
#           |_|  |_|\___/|_|  \___|_|_|\__,_|
#                             o        o     |  o
#                                 ,_       __|      ,
#                        |  |_|  /  |  |  /  |  |  / \_
#                         \/  |_/   |_/|_/\_/|_/|_/ \/
import re


from .grammar import (AST, Feature, Background, Scenario, Given, When, Then,
                      And, But, Row, Comment, Examples, Step)
from .i18n import TRANSLATIONS
from .utils import to_unicode


#  TODO  what happens with blank table items?
#  ERGO  river is to riparian as pond is to ___?

LANGUAGE_RE = re.compile(r'^# language: (\w+)')
DEFAULT_LANGUAGE = 'en'


class Parser:

    def __init__(self, language=None):
        self.thangs = [
            Feature, Background, Scenario,
            Given, When, Then, And, But,
            Row, Comment, Examples, Step
        ]
        self.steps = []
        self.language = DEFAULT_LANGUAGE if language is None else language
        self._prepare_patterns(self.language)

    def _prepare_patterns(self, language):
        self._patterns = []
        for thang in self.thangs:
            pattern = thang.get_pattern(language)
            self._patterns.append((re.compile(pattern), thang))

    def parse_file(self, filename):
        prose = open(filename, 'rb').read().decode('utf-8')
        ast = self.parse_features(prose)
        self.steps[0].filename = filename
        return ast

    def parse_features(self, prose):
        self.parse_feature(prose)
        return AST(self.steps)

    def _parse_language_directive(self, line):
        """ Parse language directive.

        :param str line: line to parse
        :returns: True if line contains correct language directive
        :side effects: sets self.language to parsed language
        """
        match = LANGUAGE_RE.match(line)
        if match:
            self.language = match.groups()[0]
            self._prepare_patterns(self.language)
            return True
        return False

    def parse_feature(self, lines):
        lines = to_unicode(lines)
        self.line_number = 0

        for self.line in lines.split(u'\n'):
            self.line_number += 1

            if not self.line:
                continue

            if self._parse_language_directive(self.line):
                continue

            if self._anneal_last_broken_line():
                continue

            if self._parse_line():
                continue

            if 0 < len(self.steps):
                self._append_to_previous_node()
            else:
                s = Step('???', self.line)
                s.line_number = self.line_number
                feature_name = TRANSLATIONS[self.language].get('feature', u'Feature')
                feature_name = feature_name.replace('|', ' or ')
                s.enforce(False, u'feature files must start with a %s' % feature_name)

        return self.steps

    def _anneal_last_broken_line(self):
        if self.steps == []:
            return False  # CONSIDER  no need me
        last_line = self.last_node.predicate

        if re.search(r'\\\s*$', last_line):
            last = self.last_node
            last.predicate += '\n' + self.line
            return True

        return False

#  TODO  permit line breakers in comments
#    | Given a table with one row
#        \| i \| be \| a \| lonely \| row |  table with only one row, line 1

    def _parse_line(self):
        self.line = self.line.rstrip()

        for regexp, klass in self._patterns:
            m = regexp.match(self.line)

            if m and len(m.groups()) > 0:
                node = klass(**m.groupdict())
                node.connect_to_parent(self.steps, self.line_number)
                self.last_node = node
                return node

    def _append_to_previous_node(self):
        previous = self.steps[-1]
        previous.predicate += '\n' + self.line.strip()
        previous.predicate = previous.predicate.strip()
        previous.validate_predicate()



# -*- coding: utf-8 -*-
import sys

from .formatters import PlainTextFormatter, ColorTextFormatter
from .parser import Parser  # noqa

__version__ = '0.4.2'


def has_color_support():
    return sys.platform != 'win32'  # pragma: nocover


def run(filename, suite, verbose=False, **kwargs):
    ''' Parse file and run tests on given suite.

    :param str filename: file name
    :param unittest.TestCase suite: TestCase instance
    :param boolean verbose: be verbose
    '''
    formatter = kwargs.get('formatter', None)
    if verbose and not formatter:
        if has_color_support():
            formatter = ColorTextFormatter()
        else:
            formatter = PlainTextFormatter()
        kwargs['formatter'] = formatter
    parser = Parser()
    ast = parser.parse_file(filename)
    return ast.evaluate(suite, **kwargs)


__all__ = ['Parser', 'run']



import unittest

from mock import sentinel, Mock, ANY

from morelia.grammar import AST


class ASTEvaluateTestCase(unittest.TestCase):
    """ Test :py:meth:`AST.evaluate`. """

    def test_should_use_provided_matcher(self):
        """ Scenariusz: matcher given as parameter """
        # Arrange
        test_visitor_class = Mock()
        matcher_class = Mock()
        feature = Mock()
        steps = [feature]
        obj = AST(steps, test_visitor_class=test_visitor_class)
        # Act
        obj.evaluate(sentinel.suite, matchers=[matcher_class])
        # Assert
        test_visitor_class.assert_called_once_with(sentinel.suite, matcher_class.return_value, ANY)

    def test_should_use_provided_formatter(self):
        """ Scenariusz: formatter given as parameter """
        # Arrange
        test_visitor = Mock()
        formatter = Mock()
        feature = Mock()
        steps = [feature]
        obj = AST(steps, test_visitor_class=test_visitor)
        # Act
        obj.evaluate(sentinel.suite, formatter=formatter)
        # Assert
        test_visitor.assert_called_once_with(sentinel.suite, ANY, formatter)

    def test_should_show_all_missing_steps(self):
        """ Scenariusz: show all missing steps """
        # Arrange
        matcher_visitor = Mock()
        feature = Mock()
        steps = [feature]
        obj = AST(steps, matcher_visitor_class=matcher_visitor)
        # Act
        obj.evaluate(sentinel.suite, show_all_missing=True)
        # Assert
        matcher_visitor.assert_called_once_with(sentinel.suite, ANY)



from abc import ABCMeta, abstractmethod
import re
import unicodedata

import parse

from .utils import to_unicode


class IStepMatcher(object):
    """ Matches methods to steps.

    Subclasses should implement at least `match` and `suggest` methods.
    """

    __metaclass__ = ABCMeta

    def __init__(self, suite, step_pattern='^step_'):
        self._suite = suite
        self._matcher = re.compile(step_pattern)
        self._next = None

    def _get_all_step_methods(self):
        match = self._matcher.match
        return [method_name for method_name in dir(self._suite) if match(method_name)]

    def add_matcher(self, matcher):
        """ Add new matcher at end of CoR

        :param matcher: matcher to add
        :returns: self
        """
        if self._next is None:
            self._next = matcher
        else:
            self._next.add_matcher(matcher)
        return self

    def find(self, predicate, augmented_predicate, step_methods=None):
        if step_methods is None:
            step_methods = self._get_all_step_methods()
        method, args, kwargs = self.match(predicate, augmented_predicate, step_methods)
        if method:
            return method, args, kwargs
        if self._next is not None:
            return self._next.find(predicate, augmented_predicate, step_methods)
        return None, (), {}

    @abstractmethod
    def match(self, predicate, augmented_predicate, step_methods):
        """ Match method from suite to given predicate.

        :param str predicate: step predicate
        :param str augmented_predicate: step augmented_predicate
        :param list step_methods: list of all step methods from suite
        :returns: (method object, args, kwargs)
        :rtype: (method, tuple, dict)
        """
        pass  # pragma: nocover

    @abstractmethod
    def suggest(self, predicate):
        """ Suggest method definition.

        Method is used to suggest methods that should be implemented.

        :param str predicate: step predicate
        :returns: (suggested method definition, suggested method name, suggested docstring)
        :rtype: (str, str, str)
        """
        pass  # pragma: nocover

    def _suggest_doc_string(self, predicate):
        predicate = predicate.replace("'", r"\'").replace('\n', r'\n')

        arguments = self._add_extra_args(r'["\<](.+?)["\>]', predicate)
        arguments = self._name_arguments(arguments)

        predicate = self.replace_placeholders(predicate, arguments)
        predicate = re.sub(r' \s+', r'\s+', predicate)

        arguments = self._format_arguments(arguments)
        return "r'%s'" % predicate, arguments

    def _name_arguments(self, extra_arguments):
        if not extra_arguments:
            return ''
        arguments = []
        number_arguments_count = sum(1 for arg_type, arg in extra_arguments
                                     if arg_type == 'number')
        if number_arguments_count < 2:
            num_suffixes = iter([''])
        else:
            num_suffixes = iter(range(1, number_arguments_count + 1))

        for arg_type, arg in extra_arguments:
            if arg_type == 'number':
                arguments.append('number%s' % next(num_suffixes))
            else:
                arguments.append(arg)
        return arguments

    def _format_arguments(self, arguments):
        if not arguments:
            return ''
        return ', ' + ', '.join(arguments)

    def replace_placeholders(self, predicate, arguments):
        predicate = re.sub(r'".+?"', '"([^"]+)"', predicate)
        predicate = re.sub(r'\<.+?\>', '(.+)', predicate)
        return predicate

    def _add_extra_args(self, matcher, predicate):
        args = re.findall(matcher, predicate)
        result = []
        for arg in args:
            try:
                float(arg)
            except ValueError:
                arg = ('id', self.slugify(arg))
            else:
                arg = ('number', arg)
            result.append(arg)
        return result

    def slugify(self, predicate):
        predicate = to_unicode(predicate)
        result = []
        for part in re.split('[^\w]+', predicate):
            part = unicodedata.normalize('NFD', part).encode('ascii', 'replace').decode('utf-8')
            part = part.replace(u'??', u'_').replace(u'?', u'')
            try:
                float(part)
            except ValueError:
                pass
            else:
                part = 'number'
            result.append(part)
        return '_'.join(result).strip('_')


class MethodNameStepMatcher(IStepMatcher):

    ''' Matcher that matches steps by method name. '''

    def match(self, predicate, augmented_predicate, step_methods):
        ''' See :py:meth:`IStepMatcher.match` '''

        clean = re.sub(r'[^\w]', '_?', predicate)
        pattern = '^step_' + clean + '$'
        regexp = re.compile(pattern)
        step_methods = [method for method in step_methods if regexp.match(method)]
        for method_name in step_methods:
            method = self._suite.__getattribute__(method_name)
            return method, (), {}
        return None, (), {}

    def suggest(self, predicate):
        ''' See :py:meth:`IStepMatcher.suggest` '''

        method_name = self.slugify(predicate)
        suggest = u'    def step_%(method_name)s(self):\n\n        raise NotImplementedError(\'%(predicate)s\')\n\n' % {
            'method_name': method_name,
            'predicate': predicate.replace("'", "\\'"),
        }
        return suggest, method_name, ''

    def slugify(self, predicate):
        predicate = to_unicode(predicate)
        predicate = unicodedata.normalize('NFD', predicate).encode('ascii', 'replace').decode('utf-8')
        predicate = predicate.replace(u'??', u'_').replace(u'?', u'')
        return re.sub(u'[^\w]+', u'_', predicate, re.U).strip('_')


class RegexpStepMatcher(IStepMatcher):

    ''' Matcher that matches steps by regexp in docstring. '''

    def match(self, predicate, augmented_predicate, step_methods):
        ''' See :py:meth:`IStepMatcher.match` '''

        for method_name in step_methods:
            method = self._suite.__getattribute__(method_name)
            doc = method.__doc__
            if not doc:
                continue
            doc = to_unicode(doc)
            doc = re.compile('^' + doc + '$')  # CONSIDER deal with users who put in the ^$
            m = doc.match(augmented_predicate)

            if m:
                kwargs = m.groupdict()
                if not kwargs:
                    args = m.groups()
                else:
                    args = ()
                return method, args, kwargs
        return None, (), {}

    def suggest(self, predicate):
        ''' See :py:meth:`IStepMatcher.suggest` '''

        docstring, extra_arguments = self._suggest_doc_string(predicate)
        method_name = self.slugify(predicate)
        suggest = u'    def step_%(method_name)s(self%(args)s):\n        %(docstring)s\n\n        raise NotImplementedError(\'%(predicate)s\')\n\n' % {
            'method_name': method_name,
            'args': extra_arguments,
            'docstring': docstring,
            'predicate': predicate.replace("'", "\\'"),
        }
        return suggest, method_name, docstring


class ParseStepMatcher(IStepMatcher):

    ''' Matcher that matches steps by format-like string in docstring. '''

    def match(self, predicate, augmented_predicate, step_methods):
        ''' See :py:meth:`IStepMatcher.match` '''

        for method_name in step_methods:
            method = self._suite.__getattribute__(method_name)
            doc = method.__doc__
            if not doc:
                continue
            match = parse.parse(doc, augmented_predicate)
            if match:
                args = match.fixed
                kwargs = match.named
                return method, tuple(args), kwargs
        return None, (), {}

    def suggest(self, predicate):
        ''' See :py:meth:`IStepMatcher.suggest` '''

        docstring, extra_arguments = self._suggest_doc_string(predicate)
        method_name = self.slugify(predicate)
        suggest = u'    def step_%(method_name)s(self%(args)s):\n        %(docstring)s\n\n        raise NotImplementedError(\'%(predicate)s\')\n\n' % {
            'method_name': method_name,
            'args': extra_arguments,
            'docstring': docstring,
            'predicate': predicate.replace("'", "\\'"),
        }
        return suggest, method_name, docstring

    def replace_placeholders(self, predicate, arguments):
        arguments = iter(arguments)

        def repl(match):
            if match.group(0).startswith('"'):
                return '"{%s}"' % next(arguments)
            return '{%s}' % next(arguments)

        predicate = re.sub(r'".+?"|\<.+?\>', repl, predicate)
        return predicate
